"""Test the houdini_package_runner.runners.pylint.runner module."""

# =============================================================================
# IMPORTS
# =============================================================================

# Standard Library
import argparse
import copy
import pathlib

# Third Party
import pytest

# Houdini Package Runner
import houdini_package_runner.config
import houdini_package_runner.items.base
import houdini_package_runner.items.dialog_script
import houdini_package_runner.items.digital_asset
import houdini_package_runner.items.filesystem
import houdini_package_runner.items.xml
import houdini_package_runner.runners.base
import houdini_package_runner.runners.pylint.runner
from houdini_package_runner.discoverers.base import BaseItemDiscoverer

# =============================================================================
# FIXTURES
# =============================================================================


@pytest.fixture
def init_runner(mocker):
    """Initialize a dummy PyLintRunner for testing."""
    mocker.patch.multiple(
        houdini_package_runner.runners.pylint.runner.PyLintRunner,
        __init__=lambda x, y: None,
    )

    def _create():
        return houdini_package_runner.runners.pylint.runner.PyLintRunner(None)

    return _create


# =============================================================================
# TESTS
# =============================================================================


class TestPyLintRunner:
    """Test houdini_package_runner.runners.pylint.runner.PyLintRunner."""

    @pytest.mark.parametrize("pass_optional", (False, True))
    def test___init__(self, mocker, pass_optional):
        """Test object initialization."""
        mock_discoverer = mocker.MagicMock(spec=BaseItemDiscoverer)

        mock_super_init = mocker.patch.object(
            houdini_package_runner.runners.pylint.runner.HoudiniPackageRunner,
            "__init__",
        )

        mock_config = (
            mocker.MagicMock(spec=houdini_package_runner.config.PackageRunnerConfig)
            if pass_optional
            else None
        )

        if pass_optional:
            inst = houdini_package_runner.runners.pylint.runner.PyLintRunner(
                mock_discoverer, runner_config=mock_config
            )
        else:
            inst = houdini_package_runner.runners.pylint.runner.PyLintRunner(
                mock_discoverer
            )

        assert inst._disabled == []

        mock_super_init.assert_called_with(mock_discoverer, runner_config=mock_config)

    # Properties

    def test_name(self, init_runner):
        """Test PyLintRunner.name."""
        inst = init_runner()

        assert inst.name == "pylint"

        with pytest.raises(AttributeError):
            inst.name = []

    # Methods

    @pytest.mark.parametrize("pass_parser", (True, False))
    def test_build_parser(self, mocker, pass_parser):
        """Test PyLintRunner.build_parser."""
        mock_parser = mocker.MagicMock(spec=argparse.ArgumentParser)

        mock_build = mocker.patch(
            "houdini_package_runner.parser.build_common_parser",
            return_value=mock_parser,
        )

        if pass_parser:
            result = (
                houdini_package_runner.runners.pylint.runner.PyLintRunner.build_parser(
                    parser=mock_parser
                )
            )

            mock_build.assert_not_called()

        else:
            result = (
                houdini_package_runner.runners.pylint.runner.PyLintRunner.build_parser()
            )

            mock_build.assert_called()

        assert result == mock_parser

        result.add_argument.assert_has_calls(
            [
                mocker.call(
                    "--rcfile",
                    action="store",
                    help="Specify a configuration file",
                ),
                mocker.call("--disable", action="store", help="Tests to disable."),
            ]
        )

    @pytest.mark.parametrize("has_options_set", (True, False))
    def test_init_args_options(self, mocker, init_runner, has_options_set):
        """Test PyLintRunner.init_args_options."""
        mock_namespace = mocker.MagicMock()

        if has_options_set:
            mock_namespace.disable = "a,b"
            mock_namespace.rcfile = "thing.rc"
            extra_args = ["arg1", "arg2"]

        else:
            mock_namespace.disable = None
            mock_namespace.rcfile = None
            extra_args = []

        expected_extra_args = copy.copy(extra_args)

        if has_options_set:
            expected_extra_args.insert(0, f"--rcfile={mock_namespace.rcfile}")

        mock_super_init = mocker.patch.object(
            houdini_package_runner.runners.pylint.runner.HoudiniPackageRunner,
            "init_args_options",
        )

        inst = init_runner()
        inst._disabled = []
        inst._ignored = []
        inst._extra_args = []

        inst.init_args_options(mock_namespace, extra_args)

        mock_super_init.assert_called_with(mock_namespace, extra_args)

        assert inst._extra_args == extra_args

        if has_options_set:
            assert inst._disabled == ["a", "b"]

        else:
            assert inst._disabled == []

    @pytest.mark.parametrize(
        "has_disabled, has_builtins, verbose",
        (
            (True, True, True),
            (False, False, False),
        ),
    )
    def test_process_path(
        self,
        mocker,
        init_runner,
        has_disabled,
        has_builtins,
        verbose,
    ):
        """Test PyLintRunner.process_path."""
        mock_io = mocker.patch("houdini_package_runner.runners.pylint.runner.StringIO")
        mock_run = mocker.patch("houdini_package_runner.runners.pylint.runner.lint.Run")
        mock_reporter = mocker.patch(
            "houdini_package_runner.runners.pylint.runner.ColorizedTextReporter"
        )

        mock_print = mocker.patch("builtins.print")
        mock_write = mocker.patch("sys.stdout.write")

        mock_colored = mocker.patch("termcolor.colored")

        mock_path = mocker.MagicMock(spec=pathlib.Path)

        mock_item = mocker.MagicMock(spec=houdini_package_runner.items.base.BaseItem)

        mock_item.ignored_builtins = ["hou"] if has_builtins else []

        mock_add_flags = mocker.patch(
            "houdini_package_runner.utils.add_or_append_to_flags"
        )

        extra_args = ["--flag", "arg"]

        mocker.patch.object(
            houdini_package_runner.runners.pylint.runner.PyLintRunner,
            "extra_args",
            extra_args,
        )

        to_ignore = (
            [
                "ABC",
            ]
            if has_disabled
            else []
        )
        extra_command = [
            "--some-flag",
        ]
        builtins = (
            [
                "abc",
            ]
            if has_builtins
            else []
        )

        mock_config = mocker.MagicMock(
            spec=houdini_package_runner.config.PackageRunnerConfig
        )
        mock_config.get_config_data.side_effect = [to_ignore, extra_command, builtins]

        mocker.patch.object(
            houdini_package_runner.runners.pylint.runner.PyLintRunner,
            "config",
            mock_config,
        )

        expected_disabled = []

        inst = init_runner()
        inst._disabled = []
        inst._extra_args = []
        inst._verbose = verbose

        if has_disabled:
            inst._disabled = ["one-thing"]
            expected_disabled = inst._disabled + to_ignore

        inst.process_path(mock_path, mock_item)

        expected_args = [str(mock_path)]
        expected_args.extend(extra_args)
        expected_args.extend(extra_command)

        assert mock_config.get_config_data.call_count == 3

        if has_builtins:
            # Only do assert_called() here as the command list will change and be inaccurate.
            mock_add_flags.assert_called()

        if expected_disabled:
            expected_args.append(f"--disable={','.join(expected_disabled)}")

        if verbose:
            mock_print.assert_has_calls(
                (
                    mocker.call(mock_colored.return_value, mock_colored.return_value),
                    mocker.call(),
                )
            )

            mock_colored.assert_has_calls(
                (
                    mocker.call(str(mock_item), "cyan"),
                    mocker.call(" ".join(expected_args[1:]), "magenta"),
                )
            )

        mock_run.assert_called_with(
            expected_args, reporter=mock_reporter.return_value, exit=False
        )

        mock_reporter.assert_called_with(mock_io.return_value)
        mock_write.assert_called_with(mock_io.return_value.getvalue.return_value)

        mock_config.get_config_data.assert_has_calls(
            [
                mocker.call("to_disable", mock_item, mock_path),
                mocker.call("command", mock_item, mock_path),
                mocker.call("known_builtins", mock_item, mock_path),
            ]
        )


def test_main(mocker):
    """Test houdini_package_runner.runners.pylint.runner.main."""
    mock_parsed = mocker.MagicMock(spec=argparse.Namespace)
    mock_unknown = mocker.MagicMock(spec=list)

    mock_parser = mocker.MagicMock(spec=argparse.ArgumentParser)
    mock_parser.parse_known_args.return_value = (mock_parsed, mock_unknown)

    mock_discoverer = mocker.MagicMock(spec=BaseItemDiscoverer)
    mock_init = mocker.patch(
        "houdini_package_runner.runners.pylint.runner.package.init_standard_package_discoverer",
        return_value=mock_discoverer,
    )

    mock_runner = mocker.MagicMock(
        spec=houdini_package_runner.runners.pylint.runner.PyLintRunner
    )

    mock_runner_init = mocker.patch(
        "houdini_package_runner.runners.pylint.runner.PyLintRunner",
        return_value=mock_runner,
    )
    mock_runner_init.build_parser.return_value = mock_parser

    result = houdini_package_runner.runners.pylint.runner.main()
    assert result == mock_runner.run.return_value

    mock_init.assert_called_with(mock_parsed)

    mock_runner_init.assert_called_with(mock_discoverer)
    mock_runner.init_args_options.assert_called_with(mock_parsed, mock_unknown)
    mock_runner.run.assert_called()



"""This module contains runnable items related to Houdini XML definitions."""

# =============================================================================
# IMPORTS
# =============================================================================

# Future
from __future__ import annotations

# Standard Library
import abc
from typing import TYPE_CHECKING, List, Tuple

# Third Party
from lxml import etree

# Houdini Package Runner
from houdini_package_runner.items.filesystem import FileToProcess

# Imports for type checking.
if TYPE_CHECKING:
    import pathlib

    from houdini_package_runner.runners.base import HoudiniPackageRunner

    ItemsToProcess = List[
        Tuple[etree._Element, str]  # pylint: disable=protected-access
    ]


# =============================================================================
# CLASSES
# =============================================================================


class XMLBase(FileToProcess, metaclass=abc.ABCMeta):
    """The base class for XML based Houdini files.

    :param path: The file path to process.
    :param write_back: Whether the item should write itself back to disk.
    :param display_name: Optional display name for test output.

    """

    def __init__(
        self, path: pathlib.Path, write_back: bool = False, display_name: str = None
    ):
        super().__init__(path, write_back=write_back, display_name=display_name)

        # hou and kwargs are always available in menu items.
        self.ignored_builtins.extend(("hou", "kwargs"))

    # -------------------------------------------------------------------------
    # NON-PUBLIC METHODS
    # -------------------------------------------------------------------------

    @abc.abstractmethod
    def _get_items_to_process(
        self,
        root: etree._Element,  # pylint: disable=protected-access
    ) -> ItemsToProcess:
        """Get any xml items that need to be processed.

        :param root: An xml element.
        :return: A list of items to be processed.

        """

    def _handle_write_back(
        self,
        section: etree._Element,  # pylint: disable=protected-access
        temp_path: pathlib.Path,
    ):
        """Handle updating the section and checking if the contents changed.

        :param section: The section to process.
        :param temp_path: The temp file name for the code.
        :return:

        """
        # Stash the pre-updated text.
        script_code = section.text

        # Read the possibly modified file.
        with temp_path.open("r") as handle:
            contents = handle.read()

        # Check if the code blob is in a CDATA block.  If the original code was in
        # a CDATA block, wrap the result in one and set it
        # back to the section.
        if "CDATA" in etree.tostring(section, encoding="unicode"):
            section.text = etree.CDATA(contents)

        # Otherwise, just use the raw values.
        else:
            section.text = contents

        if section.text != script_code:
            self.contents_changed = True

    def _process_code_section(
        self,
        section: etree._Element,  # pylint: disable=protected-access
        runner: HoudiniPackageRunner,
        base_file_name: str,
    ) -> int:
        """Process an XML section's text.

        :param section: The section to process.
        :param runner: The package runner processing the item.
        :param base_file_name: The name of the temporary file.
        :return: The process return code.

        """
        # Create a temp Python file for the code blob.
        temp_path = runner.temp_dir / f"{base_file_name}.py"

        # Dump the code to the temp file, so it can be processed.
        with temp_path.open("w") as handle:
            handle.write(str(section.text))

        result = runner.process_path(temp_path, self)

        if self.write_back:
            self._handle_write_back(section, temp_path)

        return result

    # -------------------------------------------------------------------------
    # METHODS
    # -------------------------------------------------------------------------

    def process(self, runner: HoudiniPackageRunner) -> int:
        """Process the file.

        :param runner: The package runner processing the item.
        :return: The process return code.

        """
        parser = etree.XMLParser(strip_cdata=False)
        tree = etree.parse(str(self.path), parser)

        root = tree.getroot()

        items_to_process = self._get_items_to_process(root)

        result = 0

        for item in items_to_process:
            result |= self._process_code_section(item[0], runner, item[1])

        if self.write_back and self.contents_changed:
            tree.write(str(self.path), encoding="utf-8", xml_declaration=True)

        return result


class MenuFile(XMLBase):
    """An xml menu file."""

    # -------------------------------------------------------------------------
    # NON-PUBLIC METHODS
    # -------------------------------------------------------------------------

    def _get_items_to_process(
        self,
        root: etree._Element,  # pylint: disable=protected-access
    ) -> ItemsToProcess:
        """Get any xml items that need to be processed.

        For XML menu files we look for 'scriptItem' entries and attempt to
        grab their 'scriptCode' and 'context/expression' contents.

        :param root: An xml element.
        :return: A list of items to be processed.

        """
        items = []

        for script_item in root.iter(tag="scriptItem"):
            code = script_item.find("scriptCode")

            if code is not None:
                items.append((code, str(script_item.attrib["id"])))

                context = script_item.find("context/expression")

                if context is not None:
                    items.append((context, f"{script_item.get('id'):s}.context"))

        return items


class PythonPanelFile(XMLBase):
    """A python panel file."""

    # -------------------------------------------------------------------------
    # NON-PUBLIC METHODS
    # -------------------------------------------------------------------------

    def _get_items_to_process(
        self,
        root: etree._Element,  # pylint: disable=protected-access
    ) -> ItemsToProcess:
        """Get any xml items that need to be processed.

        :param root: An xml element.
        :return: A list of items to be processed.

        """
        items = []

        for interface in root:
            script = interface.find("script")

            if script is None:
                continue

            items.append((script, str(interface.attrib["name"])))

        return items


class ShelfFile(XMLBase):
    """A shelf file.

    :param path: The shelf file path.
    :param write_back: Whether the item should write itself back to disk.
    :param display_name: Optional display name.
    :param tool_name: Optional tool name.

    """

    def __init__(
        self,
        path: pathlib.Path,
        write_back: bool = False,
        display_name: str = None,
        tool_name: str = None,
    ) -> None:
        super().__init__(path, write_back=write_back, display_name=display_name)

        self._tool_name = tool_name

    # -------------------------------------------------------------------------
    # NON-PUBLIC METHODS
    # -------------------------------------------------------------------------

    def _get_items_to_process(
        self,
        root: etree._Element,  # pylint: disable=protected-access
    ) -> ItemsToProcess:
        """Get any xml items that need to be processed.

        :param root: An xml element.
        :return: A list of items to be processed.

        """
        items = []

        for tool in root:
            script = tool.find("script")

            if script is None:
                continue

            language = script.get("scriptType")

            if language != "python":
                continue

            tool_name = str(tool.attrib["name"])

            if tool_name == "$HDA_DEFAULT_TOOL" and self._tool_name is not None:
                tool_name = (
                    self._tool_name.replace("::", "__").replace("/", "_")
                    + "_DEFAULT_TOOL"
                )

            items.append((script, tool_name))

        return items



"""Test the houdini_package_runner.items.dialog_script module."""

# =============================================================================
# IMPORTS
# =============================================================================

# Standard Library
import pathlib

# Third Party
import pytest

# Houdini Package Runner
import houdini_package_runner.items.base
import houdini_package_runner.items.dialog_script
import houdini_package_runner.runners.base

# =============================================================================
# FIXTURES
# =============================================================================


@pytest.fixture
def init_item(mocker):
    """Initialize a dummy DialogScriptItem for testing."""
    mocker.patch.multiple(
        houdini_package_runner.items.dialog_script.DialogScriptItem,
        __init__=lambda x, y, z, u, v: None,
    )

    def _create():
        return houdini_package_runner.items.dialog_script.DialogScriptItem(
            None, None, None, None
        )

    return _create


@pytest.fixture
def init_internal_item(mocker):
    """Initialize a dummy DialogScriptInternalItem for testing."""
    mocker.patch.multiple(
        houdini_package_runner.items.dialog_script.DialogScriptInternalItem,
        __init__=lambda x, y, z, u, v, w: None,
    )

    def _create():
        return houdini_package_runner.items.dialog_script.DialogScriptInternalItem(
            None, None, None, None, None
        )

    return _create


@pytest.fixture
def init_menu_item(mocker):
    """Initialize a dummy DialogScriptMenuScriptItem for testing."""
    mocker.patch.multiple(
        houdini_package_runner.items.dialog_script.DialogScriptMenuScriptItem,
        __init__=lambda x, y, z, u, v, w: None,
    )

    def _create():
        return houdini_package_runner.items.dialog_script.DialogScriptMenuScriptItem(
            None, None, None, None, None
        )

    return _create


# =============================================================================
# TESTS
# =============================================================================


class TestDialogScriptItem:
    """Test houdini_package_runner.items.dialog_script.DialogScriptItem."""

    @pytest.mark.parametrize("pass_defaults", (True, False))
    def test___init__(self, mocker, pass_defaults):
        """Test object initialization."""
        mock_path = mocker.MagicMock(spec=pathlib.Path)
        mock_name = mocker.MagicMock(spec=str)
        mock_write_back = mocker.MagicMock(spec=bool)
        mock_source_file = mocker.MagicMock(spec=pathlib.Path)

        mock_contents = mocker.MagicMock(spec=str)

        mock_handle = mocker.mock_open()
        mock_handle.return_value.read.return_value = mock_contents

        mock_path.open = mock_handle

        mock_super_init = mocker.patch.object(
            houdini_package_runner.items.base.BaseFileItem, "__init__"
        )

        mocker.patch.object(
            houdini_package_runner.items.dialog_script.DialogScriptItem,
            "path",
            new_callable=mocker.PropertyMock(return_value=mock_path),
        )

        if pass_defaults:
            inst = houdini_package_runner.items.dialog_script.DialogScriptItem(
                mock_path, mock_name
            )
            mock_super_init.assert_called_with(mock_path, write_back=False)
            assert inst._source_file is None

        else:
            inst = houdini_package_runner.items.dialog_script.DialogScriptItem(
                mock_path,
                mock_name,
                write_back=mock_write_back,
                source_file=mock_source_file,
            )
            mock_super_init.assert_called_with(mock_path, write_back=mock_write_back)
            assert inst._source_file == mock_source_file

        assert inst._name == mock_name
        assert inst._ds_contents == mock_contents

        mock_handle.assert_called_with(encoding="utf-8")

    @pytest.mark.parametrize(
        "write_back, test_file",
        (
            (False, "test__gather_items.ds"),
            (True, "test__gather_items.ds"),
            (False, None),
        ),
    )
    def test__gather_items(
        self, mocker, shared_datadir, init_item, write_back, test_file
    ):
        """Test DialogScriptItem._gather_items"""
        if test_file is not None:
            with (shared_datadir / test_file).open() as handle:
                contents = handle.read()
        else:
            contents = ""

        mock_callbacks = [mocker.MagicMock(), mocker.MagicMock()]
        mock_expressions = [mocker.MagicMock()]
        mock_menus = [mocker.MagicMock(), mocker.MagicMock()]

        for item in tuple(mock_callbacks + mock_expressions + mock_menus):
            item.write_back = False

        mock_get_callbacks = mocker.patch(
            "houdini_package_runner.items.dialog_script._get_callback_items",
            return_value=mock_callbacks,
        )
        mock_get_expressions = mocker.patch(
            "houdini_package_runner.items.dialog_script._get_expression_items",
            return_value=mock_expressions,
        )
        mock_get_menus = mocker.patch(
            "houdini_package_runner.items.dialog_script._get_menu_items",
            return_value=mock_menus,
        )

        parm_value = 'parm {\n        name    "newparameter"\n        label   "Label"\n        type    float\n        default { [ "hou.pwd().path()[-1]" python ] }\n        range   { 0 10 }\n        parmtag { "script_callback" "" }\n        parmtag { "script_callback_language" "python" }\n    }'  # noqa: E501

        mock_name = mocker.MagicMock(spec=str)

        inst = init_item()
        inst._write_back = write_back
        inst._name = mock_name
        inst._ds_contents = contents

        result = inst._gather_items()

        if test_file:
            assert result == tuple(mock_callbacks + mock_expressions + mock_menus)

            mock_get_callbacks.assert_called_with(parm_value, 4, mock_name)
            mock_get_expressions.assert_called_with(parm_value, 4, mock_name)
            mock_get_menus.assert_called_with(parm_value, 4, mock_name)

            for item in result:
                assert item.write_back == write_back

        else:
            assert not result

    def test__handle_changed_contents(self, mocker, shared_datadir, init_item):
        """Test DialogScriptItem._handle_changed_contents"""
        output_path = shared_datadir / "_handle_changed_contents.ds"

        with (
            shared_datadir / "test__handle_changed_contents" / "original.ds"
        ).open() as handle:
            original_contents = handle.read()

        with (
            shared_datadir / "test__handle_changed_contents" / "expected.ds"
        ).open() as handle:
            expected_contents = handle.read()

        mock_item1 = mocker.MagicMock()
        mock_item1.start_offset = 51
        mock_item1.end_offset = 59
        mock_item1.post_processed_code = "DIFFERENT NEW"

        mock_item2 = mocker.MagicMock()
        mock_item2.start_offset = 11
        mock_item2.end_offset = 19
        mock_item2.post_processed_code = "NEW"

        items = [mock_item1, mock_item2]

        inst = init_item()
        inst._ds_contents = original_contents
        type(inst).path = mocker.PropertyMock(return_value=output_path)

        inst._handle_changed_contents(items)

        with output_path.open() as handle:
            output_contents = handle.read()

        assert items == [mock_item2, mock_item1]
        assert output_contents == expected_contents

    # Properties

    def test_name(self, mocker, init_item):
        """Test DialogScriptItem.name"""
        mock_name = mocker.MagicMock(spec=str)

        inst = init_item()
        inst._name = mock_name

        assert inst.name == mock_name

        with pytest.raises(AttributeError):
            inst.name = "some name"

    # Methods

    @pytest.mark.parametrize(
        "write_back, has_changed_contents, return_codes, expected",
        (
            (True, True, (0, 0), 0),
            (True, False, (1, 0), 1),
            (False, True, (0, 1), 1),
        ),
    )
    def test_process(
        self,
        mocker,
        init_item,
        write_back,
        has_changed_contents,
        return_codes,
        expected,
    ):
        """Test DialogScriptItem.process."""
        mock_runner = mocker.MagicMock(
            spec=houdini_package_runner.runners.base.HoudiniPackageRunner
        )

        mock_file1 = mocker.MagicMock()
        mock_file1.contents_changed = has_changed_contents
        mock_file1.process.return_value = return_codes[0]

        mock_file2 = mocker.MagicMock()
        mock_file2.contents_changed = False
        mock_file2.process.return_value = return_codes[1]

        mocker.patch.object(
            houdini_package_runner.items.dialog_script.DialogScriptItem,
            "_gather_items",
            return_value=[mock_file1, mock_file2],
        )
        mock_handle = mocker.patch.object(
            houdini_package_runner.items.dialog_script.DialogScriptItem,
            "_handle_changed_contents",
        )

        inst = init_item()
        inst._contents_changed = False
        inst._write_back = write_back

        result = inst.process(mock_runner)

        assert result == expected

        mock_file1.process.assert_called_with(mock_runner)
        mock_file2.process.assert_called_with(mock_runner)

        if write_back and has_changed_contents:
            mock_handle.assert_called_with([mock_file1])

            assert inst.contents_changed

        else:
            assert not inst.contents_changed


class TestDialogScriptInternalItem:
    """Test houdini_package_runner.items.dialog_script.DialogScriptInternalItem."""

    @pytest.mark.parametrize("pass_defaults", (True, False))
    def test___init__(self, mocker, pass_defaults):
        """Test object initialization."""
        mock_parm = mocker.MagicMock(spec=str)
        mock_code = mocker.MagicMock(spec=str)
        mock_start_offset = mocker.MagicMock(spec=int)
        mock_end_offset = mocker.MagicMock(spec=int)
        mock_display_name = mocker.MagicMock(spec=str)
        mock_write_back = mocker.MagicMock(spec=bool)

        mock_super_init = mocker.patch.object(
            houdini_package_runner.items.dialog_script.BaseItem, "__init__"
        )

        ignored = []

        mocker.patch.object(
            houdini_package_runner.items.dialog_script.DialogScriptInternalItem,
            "ignored_builtins",
            ignored,
        )

        parm_name_result = [["first", "second"], ["third"]]
        mock_search = mocker.patch(
            "houdini_package_runner.items.dialog_script._DS_NAME_EXPR_GRAMMAR.searchString"
        )
        mock_search.return_value = parm_name_result

        if pass_defaults:
            inst = houdini_package_runner.items.dialog_script.DialogScriptInternalItem(
                mock_parm,
                mock_code,
                mock_start_offset,
                mock_end_offset,
                mock_display_name,
            )

            mock_super_init.assert_called_with(write_back=False)

        else:
            inst = houdini_package_runner.items.dialog_script.DialogScriptInternalItem(
                mock_parm,
                mock_code,
                mock_start_offset,
                mock_end_offset,
                mock_display_name,
                write_back=mock_write_back,
            )

            mock_super_init.assert_called_with(write_back=mock_write_back)

        assert inst._code == mock_code
        assert inst._end_offset == mock_end_offset
        assert inst._parm == mock_parm
        assert inst._start_offset == mock_start_offset
        assert inst._name == "first"
        assert inst._display_hint == ""
        assert inst._display_name == mock_display_name + "_" + "first"
        assert inst._post_processed_code == mock_code

        assert ignored == ["hou"]

        mock_search.assert_called_with(mock_parm)

    # Non-public methods

    @pytest.mark.parametrize(
        "code, expected_code, is_single_line",
        (
            ("foo", "foo", True),
            ("foo", "foo", False),
            ("bar\n", "bar", True),
            ("bar\n", "bar", False),
        ),
    )
    def test__load_contents(
        self, mocker, init_internal_item, code, expected_code, is_single_line
    ):
        """Test DialogScriptInternalItem._load_contents."""
        mock_temp_path = mocker.MagicMock(spec=pathlib.Path)

        mock_handle = mocker.mock_open()
        mock_handle.return_value.read.return_value = code

        mock_temp_path.open = mock_handle

        mock_escape = mocker.patch(
            "houdini_package_runner.items.dialog_script._escape_contents_for_single_line"
        )

        inst = init_internal_item()
        inst._code = code
        inst._is_single_line = is_single_line

        result = inst._load_contents(mock_temp_path)

        if is_single_line:
            assert result == mock_escape.return_value
            mock_escape.assert_called_with(expected_code)

        else:
            assert result == expected_code

    def test__post_process_contents(self, mocker, init_internal_item):
        """Test DialogScriptInternalItem._post_process_contents."""
        contents = mocker.MagicMock(spec=str)

        inst = init_internal_item()
        result = inst._post_process_contents(contents)

        assert result == contents

    @pytest.mark.parametrize(
        "code, expected_code",
        (
            ("foo", "foo\n"),
            ("bar\n", "bar\n"),
        ),
    )
    def test__write_contents(self, mocker, init_internal_item, code, expected_code):
        """Test DialogScriptInternalItem._write_contents."""
        mock_temp_path = mocker.MagicMock(spec=pathlib.Path)

        mock_handle = mocker.mock_open()

        mock_temp_path.open = mock_handle

        inst = init_internal_item()
        inst._code = code

        inst._write_contents(mock_temp_path)

        mock_handle.return_value.write.assert_called_with(expected_code)

    # Properties

    def test_code(self, mocker, init_internal_item):
        """Test DialogScriptInternalItem.code."""
        mock_code = mocker.MagicMock(spec=str)

        inst = init_internal_item()
        inst._code = mock_code

        assert inst.code == mock_code

        with pytest.raises(AttributeError):
            inst.code = "code"

    @pytest.mark.parametrize("has_hint", (True, False))
    def test_display_name(self, mocker, init_internal_item, has_hint):
        """Test DialogScriptInternalItem.display_name."""
        mock_display_name = mocker.MagicMock(spec=str)

        inst = init_internal_item()
        inst._display_name = mock_display_name
        inst._display_hint = "hint" if has_hint else ""

        if has_hint:
            assert inst.display_name == mock_display_name + "_hint"

        else:
            assert inst.display_name == mock_display_name

        with pytest.raises(AttributeError):
            inst.display_name = "display name"

    def test_end_offset(self, mocker, init_internal_item):
        """Test DialogScriptInternalItem.end_offset."""
        mock_end_offset = mocker.MagicMock(spec=int)

        inst = init_internal_item()
        inst._end_offset = mock_end_offset

        assert inst.end_offset == mock_end_offset

        with pytest.raises(AttributeError):
            inst.end_offset = 1

    def test_name(self, mocker, init_internal_item):
        """Test DialogScriptInternalItem.name."""
        mock_name = mocker.MagicMock(spec=str)

        inst = init_internal_item()
        inst._name = mock_name

        assert inst.name == mock_name

        with pytest.raises(AttributeError):
            inst.name = "name"

    def test_parm(self, mocker, init_internal_item):
        """Test DialogScriptInternalItem.parm."""
        mock_parm = mocker.MagicMock(spec=str)

        inst = init_internal_item()
        inst._parm = mock_parm

        assert inst.parm == mock_parm

        with pytest.raises(AttributeError):
            inst.parm = "parm"

    def test_start_offset(self, mocker, init_internal_item):
        """Test DialogScriptInternalItem.start_offset."""
        mock_start_offset = mocker.MagicMock(spec=int)

        inst = init_internal_item()
        inst._start_offset = mock_start_offset

        assert inst.start_offset == mock_start_offset

        with pytest.raises(AttributeError):
            inst.start_offset = 1

    def test_post_processed_code(self, mocker, init_internal_item):
        """Test DialogScriptInternalItem.post_processed_code."""
        mock_post_processed_code = mocker.MagicMock(spec=str)

        inst = init_internal_item()
        inst._post_processed_code = mock_post_processed_code

        assert inst.post_processed_code == mock_post_processed_code

        with pytest.raises(AttributeError):
            inst.post_processed_code = "code"

    # Methods

    @pytest.mark.parametrize(
        "write_back, contents_changed",
        (
            (False, False),
            (True, False),
            (True, True),
        ),
    )
    def test_process(self, mocker, init_internal_item, write_back, contents_changed):
        """Test DialogScriptInternalItem.process."""
        mock_temp_path = mocker.MagicMock(spec=pathlib.Path)

        mock_temp_dir = mocker.MagicMock(spec=pathlib.Path)
        mock_temp_dir.__truediv__.return_value = mock_temp_path

        mock_runner = mocker.MagicMock(
            spec=houdini_package_runner.runners.base.HoudiniPackageRunner
        )
        mock_runner.temp_dir = mock_temp_dir

        mock_code = mocker.MagicMock(spec=str)

        mock_contents = mocker.MagicMock(spec=str) if contents_changed else mock_code

        mock_write = mocker.patch.object(
            houdini_package_runner.items.dialog_script.DialogScriptInternalItem,
            "_write_contents",
        )

        mock_load = mocker.patch.object(
            houdini_package_runner.items.dialog_script.DialogScriptInternalItem,
            "_load_contents",
            return_value=mock_contents,
        )

        mock_post = mocker.patch.object(
            houdini_package_runner.items.dialog_script.DialogScriptInternalItem,
            "_post_process_contents",
        )

        inst = init_internal_item()
        inst._contents_changed = False
        inst._code = mock_code
        inst._post_processed_code = mock_code
        inst._write_back = write_back
        inst._display_name = "display_name"
        inst._display_hint = ""

        result = inst.process(mock_runner)

        assert result == mock_runner.process_path.return_value

        mock_temp_dir.__truediv__.assert_called_with("display_name.py")

        mock_write.assert_called_with(mock_temp_path)

        mock_runner.process_path.assert_called_with(mock_temp_path, inst)

        if write_back:
            mock_load.assert_called_with(mock_temp_path)
            if contents_changed:
                assert inst._post_processed_code == mock_post.return_value
                assert inst._contents_changed

                mock_post.assert_called_with(mock_contents)

            else:
                assert not inst._contents_changed


class TestDialogScriptCallbackItem:
    """Test houdini_package_runner.items.dialog_script.DialogScriptCallbackItem."""

    @pytest.mark.parametrize("pass_defaults", (True, False))
    def test___init__(self, mocker, pass_defaults):
        """Test object initialization."""
        mock_start_offset = mocker.MagicMock(spec=int)
        mock_end_offset = mocker.MagicMock(spec=int)

        mock_get_offset = mocker.patch(
            "houdini_package_runner.items.dialog_script._get_ds_file_offset",
            return_value=[mock_start_offset, mock_end_offset],
        )

        mock_super_init = mocker.patch.object(
            houdini_package_runner.items.dialog_script.DialogScriptInternalItem,
            "__init__",
        )

        ignored = []

        mocker.patch.object(
            houdini_package_runner.items.dialog_script.DialogScriptCallbackItem,
            "ignored_builtins",
            ignored,
        )

        mock_parm = mocker.MagicMock(spec=str)
        mock_code = mocker.MagicMock(spec=str)
        mock_parm_start = mocker.MagicMock(spec=int)
        mock_span = mocker.MagicMock(spec=tuple)
        mock_display_name = mocker.MagicMock(spec=str)
        mock_write_back = mocker.MagicMock(spec=bool)

        if pass_defaults:
            inst = houdini_package_runner.items.dialog_script.DialogScriptCallbackItem(
                mock_parm, mock_code, mock_parm_start, mock_span, mock_display_name
            )
            mock_super_init.assert_called_with(
                mock_parm,
                mock_code,
                mock_start_offset,
                mock_end_offset,
                mock_display_name,
                write_back=False,
            )

        else:
            inst = houdini_package_runner.items.dialog_script.DialogScriptCallbackItem(
                mock_parm,
                mock_code,
                mock_parm_start,
                mock_span,
                mock_display_name,
                write_back=mock_write_back,
            )
            mock_super_init.assert_called_with(
                mock_parm,
                mock_code,
                mock_start_offset,
                mock_end_offset,
                mock_display_name,
                write_back=mock_write_back,
            )

        mock_get_offset.assert_called_with(mock_parm_start, mock_span)

        assert inst._display_hint == "callback"
        assert inst._is_single_line

        assert ignored == ["kwargs"]


class TestDialogScriptDefaultExpressionItem:
    """Test houdini_package_runner.items.dialog_script.DialogScriptDefaultExpressionItem."""

    @pytest.mark.parametrize("pass_defaults", (True, False))
    def test___init__(self, mocker, pass_defaults):
        """Test object initialization."""
        mock_start_offset = mocker.MagicMock(spec=int)
        mock_end_offset = mocker.MagicMock(spec=int)

        mock_get_offset = mocker.patch(
            "houdini_package_runner.items.dialog_script._get_ds_file_offset",
            return_value=[mock_start_offset, mock_end_offset],
        )

        mock_super_init = mocker.patch.object(
            houdini_package_runner.items.dialog_script.DialogScriptInternalItem,
            "__init__",
        )

        mock_parm = mocker.MagicMock(spec=str)
        mock_code = mocker.MagicMock(spec=str)
        mock_parm_start = mocker.MagicMock(spec=int)
        mock_span = mocker.MagicMock(spec=tuple)
        mock_display_name = mocker.MagicMock(spec=str)
        mock_write_back = mocker.MagicMock(spec=bool)

        if pass_defaults:
            inst = houdini_package_runner.items.dialog_script.DialogScriptDefaultExpressionItem(
                mock_parm, mock_code, mock_parm_start, mock_span, mock_display_name
            )
            mock_super_init.assert_called_with(
                mock_parm,
                mock_code,
                mock_start_offset,
                mock_end_offset,
                mock_display_name,
                write_back=False,
            )

        else:
            inst = houdini_package_runner.items.dialog_script.DialogScriptDefaultExpressionItem(
                mock_parm,
                mock_code,
                mock_parm_start,
                mock_span,
                mock_display_name,
                write_back=mock_write_back,
            )
            mock_super_init.assert_called_with(
                mock_parm,
                mock_code,
                mock_start_offset,
                mock_end_offset,
                mock_display_name,
                write_back=mock_write_back,
            )

        mock_get_offset.assert_called_with(mock_parm_start, mock_span)

        assert inst._display_hint == "default"
        assert inst._is_single_line


class TestDialogScriptMenuScriptItem:
    """Test houdini_package_runner.items.dialog_script.DialogScriptMenuScriptItem."""

    @pytest.mark.parametrize("pass_defaults", (True, False))
    def test___init__(self, mocker, pass_defaults):
        """Test object initialization."""
        mock_start_offset = mocker.MagicMock(spec=int)
        mock_end_offset = mocker.MagicMock(spec=int)

        mock_get_offset = mocker.patch(
            "houdini_package_runner.items.dialog_script._get_ds_file_offset",
            return_value=[mock_start_offset, mock_end_offset],
        )

        mock_super_init = mocker.patch.object(
            houdini_package_runner.items.dialog_script.DialogScriptInternalItem,
            "__init__",
        )

        ignored = []

        mocker.patch.object(
            houdini_package_runner.items.dialog_script.DialogScriptMenuScriptItem,
            "ignored_builtins",
            ignored,
        )

        mock_parm = mocker.MagicMock(spec=str)
        mock_parm_start = mocker.MagicMock(spec=int)
        mock_display_name = mocker.MagicMock(spec=str)
        mock_data = mocker.MagicMock(
            spec=houdini_package_runner.items.dialog_script.PythonMenuScriptResult
        )
        mock_write_back = mocker.MagicMock(spec=bool)

        if pass_defaults:
            inst = (
                houdini_package_runner.items.dialog_script.DialogScriptMenuScriptItem(
                    mock_parm, mock_parm_start, mock_display_name, mock_data
                )
            )
            mock_super_init.assert_called_with(
                mock_parm,
                mock_data.menu_script,
                mock_start_offset,
                mock_end_offset,
                mock_display_name,
                write_back=False,
            )

        else:
            inst = (
                houdini_package_runner.items.dialog_script.DialogScriptMenuScriptItem(
                    mock_parm,
                    mock_parm_start,
                    mock_display_name,
                    mock_data,
                    write_back=mock_write_back,
                )
            )
            mock_super_init.assert_called_with(
                mock_parm,
                mock_data.menu_script,
                mock_start_offset,
                mock_end_offset,
                mock_display_name,
                write_back=mock_write_back,
            )

        mock_get_offset.assert_called_with(
            mock_parm_start, mock_data.span, inclusive=True
        )

        assert inst._display_hint == "menu_script"
        assert inst._menu_script_data == mock_data
        assert ignored == ["kwargs"]

    @pytest.mark.parametrize(
        "use_tabs, expected",
        (
            (True, '[ "foo" ]\n\t\t\t\t[ "bar" ]\n\t\t\t\t[ "baz" ]\n'),
            (False, '[ "foo" ]\n    [ "bar" ]\n    [ "baz" ]\n'),
        ),
    )
    def test__post_process_contents(self, mocker, init_menu_item, use_tabs, expected):
        """Test DialogScriptMenuScriptItem._post_process_contents."""
        mock_data = mocker.MagicMock(
            spec=houdini_package_runner.items.dialog_script.PythonMenuScriptResult
        )
        mock_data.uses_tabs = use_tabs
        mock_data.indent = 4

        contents = "foo\nbar\nbaz"

        inst = init_menu_item()
        inst._menu_script_data = mock_data

        result = inst._post_process_contents(contents)

        assert result == expected

    # Properties

    def test_menu_script_data(self, mocker, init_menu_item):
        """Test DialogScriptMenuScriptItem.menu_script_data."""
        mock_data = mocker.MagicMock(
            spec=houdini_package_runner.items.dialog_script.PythonMenuScriptResult
        )

        inst = init_menu_item()
        inst._menu_script_data = mock_data

        assert inst.menu_script_data == mock_data

        with pytest.raises(AttributeError):
            inst.menu_script_data = "some name"


@pytest.mark.parametrize(
    "value, start, expected",
    (
        ("ab\r\r\nfoo", 0, 0),
        ("\r\r\nfoo", 0, 3),
        ("\r\r\nfoo", 10, 10),
    ),
)
def test__discard_newlines(value, start, expected):
    """Test houdini_package_runner.items.dialog_script._discard_newlines."""
    result = houdini_package_runner.items.dialog_script._discard_newlines(value, start)
    assert result == expected


def test__escape_contents_for_single_line():
    """Test houdini_package_runner.items.dialog_script._escape_contents_for_single_line."""
    test_str = 'foo\rbar\n"thing"'

    result = (
        houdini_package_runner.items.dialog_script._escape_contents_for_single_line(
            test_str
        )
    )

    assert result == 'foo\\rbar\\n\\"thing\\"'


@pytest.mark.parametrize(
    "value, expected",
    (
        ('parmtag { "script_callback_language" "python" }', "python"),
        ('parmtag { "script_callback_language" "hscript" }', "hscript"),
        ("parmtag { }", None),
    ),
)
def test__get_callback_language(value, expected):
    """Test houdini_package_runner.items.dialog_script._get_callback_language."""
    result = houdini_package_runner.items.dialog_script._get_callback_language(value)

    assert result == expected


@pytest.mark.parametrize(
    "parm_start, span, inclusive, expected",
    (
        (
            (0, (3, 10), True, (3, 10)),
            (0, (3, 10), False, (4, 9)),
        )
    ),
)
def test__get_ds_file_offset(parm_start, span, inclusive, expected):
    """Test houdini_package_runner.items.dialog_script._get_ds_file_offset."""
    result = houdini_package_runner.items.dialog_script._get_ds_file_offset(
        parm_start, span, inclusive
    )

    assert result == expected


@pytest.mark.parametrize(
    "is_python, has_callback",
    (
        (True, True),
        (True, False),
        (False, False),
    ),
)
def test__get_callback_items(mocker, is_python, has_callback):
    """Test houdini_package_runner.items.dialog_script._get_callback_items."""
    mock_parm = mocker.MagicMock(spec=str)
    mock_parm_start = mocker.MagicMock(spec=int)
    mock_name = mocker.MagicMock(spec=str)

    mock_script = mocker.MagicMock(spec=str)
    mock_span = mocker.MagicMock(spec=tuple)

    language = "python" if is_python else "hscript"
    mock_get_lang = mocker.patch(
        "houdini_package_runner.items.dialog_script._get_callback_language",
        return_value=language,
    )

    script = (mock_script, mock_span) if has_callback else None

    mock_get_script = mocker.patch(
        "houdini_package_runner.items.dialog_script._get_script_callback",
        return_value=script,
    )

    mock_callback_item = mocker.patch(
        "houdini_package_runner.items.dialog_script.DialogScriptCallbackItem"
    )

    result = houdini_package_runner.items.dialog_script._get_callback_items(
        mock_parm, mock_parm_start, mock_name
    )

    if is_python:
        mock_get_script.assert_called_with(mock_parm)

    if is_python and has_callback:
        assert result == [mock_callback_item.return_value]

        mock_callback_item.assert_called_with(
            mock_parm, mock_script, mock_parm_start, mock_span, mock_name
        )

    else:
        assert not result

        mock_callback_item.assert_not_called()

    mock_get_lang.assert_called_with(mock_parm)


def test__get_default_python_expressions(shared_datadir):
    """Test houdini_package_runner.items.dialog_script._get_default_python_expressions."""
    test_path = shared_datadir / "test__get_default_python_expressions.ds"

    with test_path.open() as handle:
        contents = handle.read()

    result = houdini_package_runner.items.dialog_script._get_default_python_expressions(
        contents
    )

    assert result == (('hou.hscript("$FF")', (123, 145)),)


def test__get_expression_items(mocker):
    """Test houdini_package_runner.items.dialog_script._get_expression_items."""
    mock_parm = mocker.MagicMock(spec=str)
    mock_parm_start = mocker.MagicMock(spec=int)
    mock_name = mocker.MagicMock(spec=str)

    mock_expr = mocker.MagicMock(spec=str)
    mock_span = mocker.MagicMock(spec=tuple)

    mock_default_expressions = [(mock_expr, mock_span)]
    mock_get_exprs = mocker.patch(
        "houdini_package_runner.items.dialog_script._get_default_python_expressions",
        return_value=mock_default_expressions,
    )

    mock_expr_item = mocker.patch(
        "houdini_package_runner.items.dialog_script.DialogScriptDefaultExpressionItem"
    )

    result = houdini_package_runner.items.dialog_script._get_expression_items(
        mock_parm, mock_parm_start, mock_name
    )

    assert result == [mock_expr_item.return_value]

    mock_get_exprs.assert_called_with(mock_parm)

    mock_expr_item.assert_called_with(
        mock_parm, mock_expr, mock_parm_start, mock_span, mock_name
    )


@pytest.mark.parametrize("script_exists", (True, False))
def test__get_menu_items(mocker, script_exists):
    """Test houdini_package_runner.items.dialog_script._get_menu_items."""
    mock_parm = mocker.MagicMock(spec=str)
    mock_parm_start = mocker.MagicMock(spec=int)
    mock_name = mocker.MagicMock(spec=str)

    mock_item = (
        mocker.MagicMock(
            spec=houdini_package_runner.items.dialog_script.PythonMenuScriptResult
        )
        if script_exists
        else None
    )

    mock_get_script = mocker.patch(
        "houdini_package_runner.items.dialog_script._get_python_menu_script",
        return_value=mock_item,
    )

    mock_ds_item = mocker.patch(
        "houdini_package_runner.items.dialog_script.DialogScriptMenuScriptItem"
    )

    result = houdini_package_runner.items.dialog_script._get_menu_items(
        mock_parm, mock_parm_start, mock_name
    )

    if script_exists:
        assert result == [mock_ds_item.return_value]
        mock_ds_item.assert_called_with(
            mock_parm, mock_parm_start, mock_name, mock_item
        )

    else:
        assert not result

    mock_get_script.assert_called_with(mock_parm)


@pytest.mark.parametrize(
    "test_file, expected",
    (
        ("tabs.ds", ((84, 134), 2, True)),
        ("spaces.ds", ((129, 199), 12, False)),
        (None, ()),
    ),
)
def test__get_python_menu_script(shared_datadir, test_file, expected):
    """Test houdini_package_runner.items.dialog_script._get_python_menu_script."""
    expected_script = """import os

return [1,2,3,4]"""

    if test_file is not None:
        test_path = shared_datadir / "test__get_python_menu_script" / test_file

        with test_path.open() as handle:
            contents = handle.read()
    else:
        contents = ""

    result = houdini_package_runner.items.dialog_script._get_python_menu_script(
        contents
    )

    if test_file is not None:
        assert result.menu_script == expected_script
        assert result.span == expected[0]
        assert result.indent == expected[1]
        assert result.uses_tabs == expected[2]

    else:
        assert result is None


@pytest.mark.parametrize("has_match", (True, False))
def test__get_script_callback(shared_datadir, has_match):
    """Test houdini_package_runner.items.dialog_script._get_script_callback."""

    if has_match:
        test_path = shared_datadir / "test__get_script_callback.ds"

        with test_path.open() as handle:
            contents = handle.read()
    else:
        contents = ""

    result = houdini_package_runner.items.dialog_script._get_script_callback(contents)

    if has_match:
        assert result == ("hou.hm().callback(hou.pwd())", (165, 195))

    else:
        assert result is None



"""setup.py for houdini_package_runner."""
# Standard Library
from pathlib import Path

# Third Party
from setuptools import find_packages, setup

# Package meta-data.
NAME = "houdini_package_runner"
DESCRIPTION = "Tooling to analyze or perform actions on Python code in Houdini packages"
URL = "https://github.com/captainhammy/houdini_package_runner"
AUTHOR = "Graham Thompson"
AUTHOR_EMAIL = "captainhammy@gmail.com"
REQUIRES_PYTHON = ">=3.7.0"

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name=NAME,
    use_scm_version=True,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    package_dir={"": "src"},
    packages=find_packages(where="src", exclude=("tests",)),
    install_requires=[
        "black",
        "deepmerge",
        "flake8",
        "isort",
        "lxml",
        "modernize",
        "pylint",
        "pyparsing",
        "termcolor",
        "toml",
    ],
    extras_require={
        "test": [
            "pytest",
            "pytest-cov",
            "pytest-datadir",
            "pytest-mock",
            "pytest-subprocess",
            "tox",
        ]
    },
    entry_points={
        "console_scripts": [
            "houdini_package_black=houdini_package_runner.runners.black.runner:main",
            "houdini_package_flake8=houdini_package_runner.runners.flake8.runner:main",
            "houdini_package_isort=houdini_package_runner.runners.isort.runner:main",
            "houdini_package_modernize=houdini_package_runner.runners.modernize.runner:main",
            "houdini_package_pylint=houdini_package_runner.runners.pylint.runner:main",
        ],
    },
    include_package_data=True,
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Software Development :: Testing",
        "Topic :: Utilities",
    ],
)



"""Test the houdini_package_runner.items.filesystem module."""

# =============================================================================
# IMPORTS
# =============================================================================

# Standard Library
import pathlib

# Third Party
import pytest

# Houdini Package Runner
import houdini_package_runner.items.base
import houdini_package_runner.items.filesystem
import houdini_package_runner.runners.base

# =============================================================================
# FIXTURES
# =============================================================================


@pytest.fixture
def init_dir_item(mocker):
    """Initialize a dummy DirectoryToProcess for testing."""
    mocker.patch.multiple(
        houdini_package_runner.items.filesystem.DirectoryToProcess,
        __init__=lambda x, y, z, w: None,
    )

    def _create():
        return houdini_package_runner.items.filesystem.DirectoryToProcess(
            None, None, None
        )

    return _create


@pytest.fixture
def init_file_item(mocker):
    """Initialize a dummy FileToProcess for testing."""
    mocker.patch.multiple(
        houdini_package_runner.items.filesystem.FileToProcess,
        __init__=lambda x, y, z, w: None,
    )

    def _create():
        return houdini_package_runner.items.filesystem.FileToProcess(None, None, None)

    return _create


@pytest.fixture
def init_houdini_scripts_dir_item(mocker):
    """Initialize a dummy HoudiniScriptsDirectoryItem for testing."""
    mocker.patch.multiple(
        houdini_package_runner.items.filesystem.HoudiniScriptsDirectoryItem,
        __init__=lambda x, y, z, w: None,
    )

    def _create():
        return houdini_package_runner.items.filesystem.HoudiniScriptsDirectoryItem(
            None, None, None
        )

    return _create


# =============================================================================
# TESTS
# =============================================================================


class TestDirectoryToProcess:
    """Test houdini_package_runner.base.filesystem.DirectoryToProcess."""

    @pytest.mark.parametrize("traverse", (None, True, False))
    def test___init__(self, mocker, traverse):
        """Test object initialization."""
        mock_path = mocker.MagicMock(spec=pathlib.Path)
        mock_write_back = mocker.MagicMock(spec=bool)

        mock_super_init = mocker.patch.object(
            houdini_package_runner.items.base.BaseFileItem, "__init__"
        )

        if traverse is not None:
            inst = houdini_package_runner.items.filesystem.DirectoryToProcess(
                mock_path, write_back=mock_write_back, traverse_children=traverse
            )

        else:
            inst = houdini_package_runner.items.filesystem.DirectoryToProcess(
                mock_path, write_back=mock_write_back
            )

        mock_super_init.assert_called_with(mock_path, write_back=mock_write_back)
        assert inst._traverse_children == bool(traverse)

    # Non-Public Methods

    @pytest.mark.parametrize("is_test", (True, False))
    def test__get_child_items(self, shared_datadir, is_test):
        """Test DirectoryToProcess._get_child_items."""
        dir_path = shared_datadir / "test_get_child_items"

        inst = houdini_package_runner.items.filesystem.DirectoryToProcess(dir_path)
        inst.is_test_item = is_test
        result = inst._get_child_items()

        assert len(result) == 3

        if is_test:
            for item in result:
                assert item.is_test_item

    @pytest.mark.parametrize(
        "return_codes, expected",
        (
            ((0, 0), 0),
            ((1, 0), 1),
            ((0, 1), 1),
        ),
    )
    def test__process_children(self, mocker, init_dir_item, return_codes, expected):
        """Test DirectoryToProcess._process_children."""
        mock_runner = mocker.MagicMock(
            spec=houdini_package_runner.runners.base.HoudiniPackageRunner
        )

        mock_file = mocker.MagicMock(
            spec=houdini_package_runner.items.filesystem.FileToProcess
        )
        mock_file.ignored_builtins = []
        mock_file.process.return_value = return_codes[0]

        mock_dir = mocker.MagicMock(
            spec=houdini_package_runner.items.filesystem.DirectoryToProcess
        )
        mock_dir.ignored_builtins = []
        mock_dir.process.return_value = return_codes[1]

        mock_get_items = mocker.patch.object(
            houdini_package_runner.items.filesystem.DirectoryToProcess,
            "_get_child_items",
            return_value=(mock_dir, mock_file),
        )

        inst = init_dir_item()

        result = inst._process_children(mock_runner)
        assert result == expected

        mock_get_items.assert_called()

        mock_file.process.assert_called_with(mock_runner)
        mock_dir.process.assert_called_with(mock_runner)

    # Properties

    def test_traverse_children(self, mocker, init_dir_item):
        """Test DirectoryToProcess.traverse_children."""
        mock_traverse = mocker.MagicMock(spec=bool)
        inst = init_dir_item()
        inst._traverse_children = mock_traverse

        assert inst.traverse_children == mock_traverse

    # Methods

    @pytest.mark.parametrize("traverse", (True, False))
    def test_process(self, mocker, init_dir_item, traverse):
        """Test DirectoryToProcess.process."""
        mock_runner = mocker.MagicMock(
            spec=houdini_package_runner.runners.base.HoudiniPackageRunner
        )

        mock_path = mocker.MagicMock(spec=pathlib.Path)

        inst = init_dir_item()
        mocker.patch.object(
            houdini_package_runner.items.filesystem.DirectoryToProcess,
            "path",
            mocker.PropertyMock(return_value=mock_path),
        )

        mocker.patch.object(
            houdini_package_runner.items.filesystem.DirectoryToProcess,
            "traverse_children",
            mocker.PropertyMock(return_value=traverse),
        )

        mock_process = mocker.patch.object(
            houdini_package_runner.items.filesystem.DirectoryToProcess,
            "_process_children",
        )

        result = inst.process(mock_runner)

        if traverse:
            assert result == mock_process.return_value
            mock_process.assert_called_with(mock_runner)
        else:
            assert result == mock_runner.process_path.return_value
            mock_runner.process_path.assert_called_with(mock_path, inst)


class TestFileToProcess:
    """Test houdini_package_runner.base.filesystem.FileToProcess."""

    @pytest.mark.parametrize("display_name", (None, "display_name"))
    def test___init__(self, mocker, display_name):
        """Test object initialization."""
        mock_path = mocker.MagicMock(spec=pathlib.Path)
        mock_write_back = mocker.MagicMock(spec=bool)

        mock_super_init = mocker.patch.object(
            houdini_package_runner.items.base.BaseFileItem, "__init__"
        )

        if display_name is not None:
            inst = houdini_package_runner.items.filesystem.FileToProcess(
                mock_path, write_back=mock_write_back, display_name=display_name
            )

        else:
            inst = houdini_package_runner.items.filesystem.FileToProcess(
                mock_path, write_back=mock_write_back
            )

        mock_super_init.assert_called_with(mock_path, write_back=mock_write_back)
        assert inst._display_name == display_name

    # Properties

    def test_display_name(self, mocker, init_file_item):
        """Test FileToProcess.display_name."""
        mock_display_name = mocker.MagicMock(spec=str)

        inst = init_file_item()
        inst._display_name = mock_display_name

        assert inst.display_name == mock_display_name

        mock_set_display_name = mocker.MagicMock(spec=str)
        inst.display_name = mock_set_display_name

        assert inst._display_name == mock_set_display_name

    # Methods

    @pytest.mark.parametrize("contents_changed", (True, False))
    def test_process(self, mocker, init_file_item, contents_changed):
        """Test FileToProcess.process."""
        mock_runner = mocker.MagicMock(
            spec=houdini_package_runner.runners.base.HoudiniPackageRunner
        )

        mock_hash1 = mocker.MagicMock(spec=bytes)
        mock_hash2 = mocker.MagicMock(spec=bytes)

        results = (
            (mock_hash1, mock_hash2) if contents_changed else (mock_hash1, mock_hash1)
        )

        mock_compute = mocker.patch(
            "houdini_package_runner.items.filesystem.compute_file_hash",
            side_effect=results,
        )

        mock_path = mocker.MagicMock(spec=pathlib.Path)

        inst = init_file_item()
        mocker.patch.object(
            houdini_package_runner.items.filesystem.FileToProcess,
            "path",
            mocker.PropertyMock(return_value=mock_path),
        )

        result = inst.process(mock_runner)

        assert result == mock_runner.process_path.return_value

        assert inst.contents_changed == contents_changed

        mock_runner.process_path.assert_called_with(mock_path, inst)
        mock_compute.assert_has_calls(
            [
                mocker.call(mock_path),
                mocker.call(mock_path),
            ]
        )


class TestHoudiniScriptsDirectoryItem:
    """Test houdini_package_runner.base.filesystem.HoudiniScriptsDirectoryItem."""

    def test__get_child_items(self, mocker, init_houdini_scripts_dir_item):
        """Test HoudiniScriptsDirectoryItem._get_child_items."""
        mock_file = mocker.MagicMock(
            spec=houdini_package_runner.items.filesystem.FileToProcess
        )
        mock_file.ignored_builtins = []

        mock_dir = mocker.MagicMock(
            spec=houdini_package_runner.items.filesystem.DirectoryToProcess
        )
        mock_dir.ignored_builtins = []

        mock_super_get = mocker.patch.object(
            houdini_package_runner.items.filesystem.DirectoryToProcess,
            "_get_child_items",
            return_value=(mock_dir, mock_file),
        )

        inst = init_houdini_scripts_dir_item()

        inst._get_child_items()

        mock_super_get.assert_called()
        assert mock_file.ignored_builtins == ["kwargs"]
        assert mock_dir.ignored_builtins == []


class TestPythonPackageDirectoryItem:
    """Test houdini_package_runner.base.filesystem.PythonPackageDirectoryItem."""

    def test___init__(self, mocker):
        """Test object initialization."""
        mock_path = mocker.MagicMock(spec=pathlib.Path)
        mock_write_back = mocker.MagicMock(spec=bool)

        mock_super_init = mocker.patch.object(
            houdini_package_runner.items.filesystem.DirectoryToProcess, "__init__"
        )

        houdini_package_runner.items.filesystem.PythonPackageDirectoryItem(
            mock_path, write_back=mock_write_back
        )

        mock_super_init.assert_called_with(mock_path, write_back=mock_write_back)


@pytest.mark.parametrize(
    "file_name, expected",
    [
        ("python_file.py", "d41d8cd98f00b204e9800998ecf8427e"),
        ("decode_error.hip", "7db20f4b9bff4b5b07d067ed70521b71"),
    ],
)
def test_compute_file_hash(shared_datadir, file_name, expected):
    """Test houdini_package_runner.items.filesystem.compute_file_hash."""
    file_path = shared_datadir / "test_is_python_file" / file_name

    result = houdini_package_runner.items.filesystem.compute_file_hash(file_path)

    assert result == expected


@pytest.mark.parametrize(
    "file_name, bins, expected",
    [
        ("python_file.py", None, True),
        ("python_file.pyc", None, False),
        ("ls_script", None, False),
        ("no_shbang_script", None, False),
        ("decode_error.hip", None, False),
        ("python_script", None, True),
        ("hython_script", None, False),
        ("hython_script", ["hython"], True),
        ("hython_script", ["python", "hython"], True),
    ],
)
def test_is_python_file(shared_datadir, file_name, bins, expected):
    """Test houdini_package_runner.items.filesystem.is_python_file."""
    file_path = shared_datadir / "test_is_python_file" / file_name
    result = houdini_package_runner.items.filesystem.is_python_file(
        file_path, python_bins=bins
    )

    assert result == expected



"""This module contains runnable items related to entities on disk."""

# =============================================================================
# IMPORTS
# =============================================================================

# Future
from __future__ import annotations

# Standard Library
import hashlib
import pathlib
from typing import TYPE_CHECKING, List, Optional

# Houdini Package Runner
from houdini_package_runner.items.base import BaseFileItem

# Imports for type checking.
if TYPE_CHECKING:
    import houdini_package_runner.runners.base
    from houdini_package_runner.items.base import BaseItem


# =============================================================================
# CLASSES
# =============================================================================


class DirectoryToProcess(BaseFileItem):
    """Class representing a directory to be processed.

    :param path: The path for the item.
    :param write_back: Whether the item should write itself back to disk.
    :param traverse_children: Whether to traverse a directories children.

    """

    def __init__(
        self,
        path: pathlib.Path,
        write_back: bool = False,
        traverse_children: bool = False,
    ) -> None:
        super().__init__(path, write_back=write_back)

        self._traverse_children = traverse_children

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.path} traverse_children={self.traverse_children}>"

    # -------------------------------------------------------------------------
    # NON-PUBLIC METHODS
    # -------------------------------------------------------------------------

    def _get_child_items(self) -> List[BaseItem]:
        """Find child items to process.

        :return: A list of found child items.

        """
        items: List[BaseItem] = []

        for child in pathlib.Path(self.path).iterdir():
            # Skip hidden items.
            if not child.name[0].isalpha():
                continue

            if child.is_dir():
                # If there is an __init__.py file then we can treat the directory as a
                # Python package.
                if (child / "__init__.py").exists():
                    items.append(
                        PythonPackageDirectoryItem(child, write_back=self.write_back)
                    )

                # Otherwise, it's just another directory.
                else:
                    items.append(
                        DirectoryToProcess(
                            child, write_back=self.write_back, traverse_children=True
                        )
                    )

            else:
                # If the file is a Python file we can process it.
                if is_python_file(child):
                    items.append(FileToProcess(child, write_back=self.write_back))

        # If this is a test item then make any found child items also test items.
        if self.is_test_item:
            for item in items:
                item.is_test_item = True

        return items

    def _process_children(
        self, runner: houdini_package_runner.runners.base.HoudiniPackageRunner
    ) -> int:
        """Process an item.

        :param runner: The package runner processing the item.
        :return: The process return code.

        """
        items = self._get_child_items()

        result = 0

        # Process each item with the runner.
        for item in items:
            result |= item.process(runner)

        return result

    # -------------------------------------------------------------------------
    # PROPERTIES
    # -------------------------------------------------------------------------

    @property
    def traverse_children(self) -> bool:
        """Whether to traverse children."""
        return self._traverse_children

    # -------------------------------------------------------------------------
    # METHODS
    # -------------------------------------------------------------------------

    def process(
        self, runner: houdini_package_runner.runners.base.HoudiniPackageRunner
    ) -> int:
        """Process an item.

        :param runner: The package runner processing the item.
        :return: The process return code.

        """
        # If traversing children then we want to process them.
        if self.traverse_children:
            return self._process_children(runner)

        # Otherwise, process this directory.
        return runner.process_path(self.path, self)


class FileToProcess(BaseFileItem):
    """Class representing a file to process.

    :param path: The file path to process.
    :param write_back: Whether the item should write itself back to disk.
    :param display_name: Optional display name for test output.

    """

    def __init__(
        self, path: pathlib.Path, write_back: bool = False, display_name: str = None
    ) -> None:
        super().__init__(path, write_back=write_back)

        self._display_name = display_name

    def __repr__(self):
        path = self.display_name if self.display_name is not None else self.path

        return f"<{self.__class__.__name__} {path}>"

    # -------------------------------------------------------------------------
    # PROPERTIES
    # -------------------------------------------------------------------------

    @property
    def display_name(self) -> Optional[str]:
        """Display name for test output."""
        return self._display_name

    @display_name.setter
    def display_name(self, display_name: str):
        self._display_name = display_name

    # -------------------------------------------------------------------------
    # METHODS
    # -------------------------------------------------------------------------

    def process(
        self, runner: houdini_package_runner.runners.base.HoudiniPackageRunner
    ) -> int:
        """Process the file.

        :param runner: The package runner processing the item.
        :return: The process return code.

        """
        pre_hash = compute_file_hash(self.path)
        result = runner.process_path(self.path, self)
        post_hash = compute_file_hash(self.path)
        self.contents_changed = pre_hash != post_hash

        return result


class HoudiniDirectoryItem(DirectoryToProcess):
    """Subclass to represent a directory under a package houdini/ directory."""


class HoudiniScriptsDirectoryItem(DirectoryToProcess):
    """Subclass to represent a scripts/ directory under a package houdini directory."""

    def _get_child_items(self) -> List[BaseItem]:
        """Find child items to process.

        :return: A list of found child items.

        """
        items = super()._get_child_items()

        for item in items:
            # Mark any child files is ignoring the 'kwargs' builtin as it will always
            # exist in the handler scripts.
            if isinstance(item, FileToProcess):
                item.ignored_builtins.append("kwargs")

        return items


class PythonPackageDirectoryItem(DirectoryToProcess):
    """Subclass to represent a directory which is a Python package."""

    def __init__(self, path: pathlib.Path, write_back: bool = False) -> None:
        super().__init__(path, write_back=write_back)

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.path}>"


# =============================================================================
# FUNCTIONS
# =============================================================================


def compute_file_hash(file_path: pathlib.Path) -> str:
    """Compute a hash for the file contents.

    :param file_path: The path to the file to hash.
    :return: The computed file hash.

    """
    with open(file_path, "rb") as handle:
        file_hash = hashlib.md5()
        # TODO: Replace with walrus operator once we commit to Python 3.8+.
        # while chunk := handle.read(8192):
        chunk = handle.read(8192)
        while chunk:
            file_hash.update(chunk)
            chunk = handle.read(8192)

    return file_hash.hexdigest()


def is_python_file(
    file_path: pathlib.Path, python_bins: Optional[List[str]] = None
) -> bool:
    """Check if a file is a Python file.

    This function will do it's best to determine Whether a file is a Python file.

    It will:
      - Check if the extension is .py
      - Check if the first line is a #! line and Whether any of the python_bin names are in the line.

    :param file_path: The path to check.
    :param python_bins: Optional list of python executable names.
    :return: Whether this is a Python file.

    """
    ext = file_path.suffix

    if ext == ".py":
        return True

    if ext == ".pyc":
        return False

    # Try to peak at the first line.
    with file_path.open() as handle:
        try:
            first_line = handle.readline()

        except UnicodeDecodeError:
            return False

    python_bins = ["python"] if python_bins is None else python_bins

    # If the file is a script check if any of the python bin names are in the command.
    if first_line.startswith("#!"):
        for python_bin in python_bins:
            if python_bin in first_line:
                return True

    return False



"""Test the houdini_package_runner.runners.black.runner module."""

# =============================================================================
# IMPORTS
# =============================================================================

# Standard Library
import argparse
import pathlib

# Third Party
import pytest

# Houdini Package Runner
import houdini_package_runner.config
import houdini_package_runner.items.base
import houdini_package_runner.items.dialog_script
import houdini_package_runner.items.digital_asset
import houdini_package_runner.items.filesystem
import houdini_package_runner.items.xml
import houdini_package_runner.runners.base
import houdini_package_runner.runners.black.runner
from houdini_package_runner.discoverers.base import BaseItemDiscoverer

# =============================================================================
# FIXTURES
# =============================================================================


@pytest.fixture
def init_runner(mocker):
    """Initialize a dummy BlackRunner for testing."""
    mocker.patch.multiple(
        houdini_package_runner.runners.black.runner.BlackRunner,
        __init__=lambda x, y: None,
    )

    def _create():
        return houdini_package_runner.runners.black.runner.BlackRunner(None)

    return _create


# =============================================================================
# TESTS
# =============================================================================


class TestBlackRunner:
    """Test houdini_package_runner.runners.black.runner.BlackRunner."""

    # Properties

    def test_name(self, init_runner):
        """Test BlackRunner.name."""
        inst = init_runner()

        assert inst.name == "black"

        with pytest.raises(AttributeError):
            inst.name = []

    # Methods

    @pytest.mark.parametrize("pass_parser", (True, False))
    def test_build_parser(self, mocker, pass_parser):
        """Test BlackRunner.build_parser."""
        mock_build = mocker.patch("houdini_package_runner.parser.build_common_parser")

        if pass_parser:
            mock_parser = mocker.MagicMock(spec=argparse.ArgumentParser)

            result = (
                houdini_package_runner.runners.black.runner.BlackRunner.build_parser(
                    parser=mock_parser
                )
            )
            assert result == mock_parser

            mock_build.assert_not_called()

        else:
            result = (
                houdini_package_runner.runners.black.runner.BlackRunner.build_parser()
            )

            assert result == mock_build.return_value

    def test_init_args_options(self, mocker, init_runner):
        """Test BlackRunner.init_args_options."""
        mock_namespace = mocker.MagicMock(spec=argparse.Namespace)
        mock_extra_args = mocker.MagicMock(spec=list)

        mock_super_init = mocker.patch.object(
            houdini_package_runner.runners.black.runner.HoudiniPackageRunner,
            "init_args_options",
        )

        inst = init_runner()

        inst.init_args_options(mock_namespace, mock_extra_args)

        mock_super_init.assert_called_with(mock_namespace, mock_extra_args)

        assert inst._extra_args == mock_extra_args

    @pytest.mark.parametrize("pass_target_version", (True, False))
    def test_process_path(self, mocker, init_runner, pass_target_version):
        """Test BlackRunner.process_path."""
        mock_path = mocker.MagicMock(spec=pathlib.Path)

        mock_item = mocker.MagicMock(spec=houdini_package_runner.items.base.BaseItem)

        mock_config = mocker.MagicMock(
            spec=houdini_package_runner.config.PackageRunnerConfig
        )
        mock_config.get_config_data.return_value = ["--foo", "bar"]

        mocker.patch.object(
            houdini_package_runner.runners.black.runner.BlackRunner,
            "config",
            mock_config,
        )

        mock_execute = mocker.patch(
            "houdini_package_runner.utils.execute_subprocess_command"
        )

        extra_args = ["--flag", "arg"]

        if pass_target_version:
            extra_args.append("--target-version=py34")

        mocker.patch.object(
            houdini_package_runner.runners.black.runner.BlackRunner,
            "extra_args",
            extra_args,
        )

        mock_verbose = mocker.MagicMock(spec=bool)

        inst = init_runner()
        inst._verbose = mock_verbose

        inst.process_path(mock_path, mock_item)

        mock_config.get_config_data.assert_called_with("flags", mock_item, mock_path)

        expected_args = ["black"] + ["--foo", "bar"] + extra_args + [str(mock_path)]

        if not pass_target_version:
            expected_args.insert(1, "--target-version=py37")

        mock_execute.assert_called_with(expected_args, verbose=mock_verbose)


def test_main(mocker):
    """Test houdini_package_runner.runners.black.runner.main."""
    mock_parsed = mocker.MagicMock(spec=argparse.Namespace)
    mock_unknown = mocker.MagicMock(spec=list)

    mock_parser = mocker.MagicMock(spec=argparse.ArgumentParser)
    mock_parser.parse_known_args.return_value = (mock_parsed, mock_unknown)

    mock_discoverer = mocker.MagicMock(spec=BaseItemDiscoverer)
    mock_init = mocker.patch(
        "houdini_package_runner.runners.black.runner.package.init_standard_package_discoverer",
        return_value=mock_discoverer,
    )

    mock_runner = mocker.MagicMock(
        spec=houdini_package_runner.runners.black.runner.BlackRunner
    )

    mock_runner_init = mocker.patch(
        "houdini_package_runner.runners.black.runner.BlackRunner",
        return_value=mock_runner,
    )
    mock_runner_init.build_parser.return_value = mock_parser

    result = houdini_package_runner.runners.black.runner.main()
    assert result == mock_runner.run.return_value

    mock_init.assert_called_with(mock_parsed)

    mock_runner_init.assert_called_with(mock_discoverer)
    mock_runner.init_args_options.assert_called_with(mock_parsed, mock_unknown)
    mock_runner.run.assert_called()



"""Test the houdini_package_runner.items.digital_asset module."""

# =============================================================================
# IMPORTS
# =============================================================================

# Standard Library
import pathlib

# Third Party
import pytest

# Houdini Package Runner
import houdini_package_runner.items.base
import houdini_package_runner.items.dialog_script
import houdini_package_runner.items.digital_asset
import houdini_package_runner.items.filesystem
import houdini_package_runner.items.xml
import houdini_package_runner.runners.base

# =============================================================================
# FIXTURES
# =============================================================================


@pytest.fixture
def init_expanded(mocker):
    """Initialize a dummy ExpandedOperatorType for testing."""
    mocker.patch.multiple(
        houdini_package_runner.items.digital_asset.ExpandedOperatorType,
        __init__=lambda x, y, z, u, v: None,
    )

    def _create():
        return houdini_package_runner.items.digital_asset.ExpandedOperatorType(
            None, None, None, None
        )

    return _create


@pytest.fixture
def init_asset_dir(mocker):
    """Initialize a dummy DigitalAssetDirectory for testing."""
    mocker.patch.multiple(
        houdini_package_runner.items.digital_asset.DigitalAssetDirectory,
        __init__=lambda x, y, z, u: None,
    )

    def _create():
        return houdini_package_runner.items.digital_asset.DigitalAssetDirectory(
            None, None, None
        )

    return _create


@pytest.fixture
def init_binary(mocker):
    """Initialize a dummy BinaryDigitalAssetFile for testing."""
    mocker.patch.multiple(
        houdini_package_runner.items.digital_asset.BinaryDigitalAssetFile,
        __init__=lambda x, y, z: None,
    )

    def _create():
        return houdini_package_runner.items.digital_asset.BinaryDigitalAssetFile(
            None, None
        )

    return _create


# =============================================================================
# TESTS
# =============================================================================


class TestDigitalAssetPythonSection:
    """Test houdini_package_runner.items.digital_asset.DigitalAssetPythonSection."""

    @pytest.mark.parametrize(
        "section_name, write_back",
        (
            ("OnCreated", False),
            ("PythonModule", True),
            ("PythonCook", True),
        ),
    )
    def test___init__(self, mocker, section_name, write_back):
        """Test object initialization."""
        mock_path = mocker.MagicMock(spec=pathlib.Path)
        name = "/some/path/" + section_name

        if write_back:
            inst = houdini_package_runner.items.digital_asset.DigitalAssetPythonSection(
                mock_path,
                name,
                write_back=write_back,
            )

        else:
            inst = houdini_package_runner.items.digital_asset.DigitalAssetPythonSection(
                mock_path, name
            )

        if section_name not in ("PythonCook", "PythonModule"):
            assert "kwargs" in inst.ignored_builtins


class TestExpandedOperatorType:
    """Test houdini_package_runner.items.digital_asset.ExpandedOperatorType."""

    @pytest.mark.parametrize("write_back", (False, True))
    def test___init__(self, mocker, write_back):
        """Test object initialization."""
        mock_path = mocker.MagicMock(spec=pathlib.Path)
        mock_name = mocker.MagicMock(spec=str)
        mock_source_file = mocker.MagicMock(spec=pathlib.PurePath)

        mock_super_init = mocker.patch.object(
            houdini_package_runner.items.base.BaseFileItem, "__init__"
        )

        if write_back:
            inst = houdini_package_runner.items.digital_asset.ExpandedOperatorType(
                mock_path,
                mock_name,
                write_back=write_back,
                source_file=mock_source_file,
            )
            assert inst._source_file == mock_source_file

        else:
            inst = houdini_package_runner.items.digital_asset.ExpandedOperatorType(
                mock_path, mock_name
            )
            assert inst._source_file is None

        mock_super_init.assert_called_with(mock_path, write_back=write_back)

        assert inst._name == mock_name

    # Non-Public Methods

    @pytest.mark.parametrize("has_sourcefile", (True, False))
    def test__build_python_section_items(self, mocker, init_expanded, has_sourcefile):
        """Test ExpandedOperatorType._build_python_section_items."""
        mock_file1 = mocker.MagicMock(
            spec=houdini_package_runner.items.digital_asset.DigitalAssetPythonSection
        )
        mock_file2 = mocker.MagicMock(
            spec=houdini_package_runner.items.digital_asset.DigitalAssetPythonSection
        )

        mock_init_file = mocker.patch(
            "houdini_package_runner.items.digital_asset.DigitalAssetPythonSection",
            side_effect=[mock_file1, mock_file2],
        )

        mock_source_file = mocker.MagicMock(spec=pathlib.Path)

        mock_path = mocker.MagicMock(spec=pathlib.Path)

        mock_write_back = mocker.MagicMock(spec=bool)

        mocker.patch.object(
            houdini_package_runner.items.digital_asset.ExpandedOperatorType,
            "path",
            mock_path,
        )
        mocker.patch.object(
            houdini_package_runner.items.digital_asset.ExpandedOperatorType,
            "write_back",
            mock_write_back,
        )

        mock_section1 = mocker.MagicMock(spec=pathlib.Path)
        mock_section2 = mocker.MagicMock(spec=pathlib.Path)

        sections = [mock_section1, mock_section2]

        inst = init_expanded()
        inst._source_file = mock_source_file if has_sourcefile else None

        if has_sourcefile:
            expected_display_name1 = mock_source_file / mock_section1.name
            expected_display_name2 = mock_source_file / mock_section2.name

        else:
            expected_display_name1 = mock_path / mock_section1.name
            expected_display_name2 = mock_path / mock_section2.name

        result = inst._build_python_section_items(sections)

        assert result == [mock_file1, mock_file2]

        mock_init_file.assert_has_calls(
            [
                mocker.call(
                    mock_section1,
                    str(expected_display_name1),
                    write_back=mock_write_back,
                ),
                mocker.call(
                    mock_section2,
                    str(expected_display_name2),
                    write_back=mock_write_back,
                ),
            ]
        )

    @pytest.mark.parametrize("cook_exists", (True, False))
    def test__find_python_section_paths(self, mocker, init_expanded, cook_exists):
        """Test ExpandedOperatorType._find_python_sections."""
        mock_section1 = mocker.MagicMock(spec=pathlib.Path)
        mock_section2 = mocker.MagicMock(spec=pathlib.Path)

        mocker.patch.object(
            houdini_package_runner.items.digital_asset.ExpandedOperatorType,
            "_get_extra_python_section_files",
            return_value=[mock_section1, mock_section2],
        )

        mock_path = mocker.MagicMock(spec=pathlib.Path)
        mock_path.__truediv__.return_value.exists.return_value = cook_exists

        mocker.patch.object(
            houdini_package_runner.items.digital_asset.ExpandedOperatorType,
            "path",
            mock_path,
        )

        expected = [mock_section1, mock_section2]

        if cook_exists:
            expected.append(mock_path.__truediv__.return_value)

        inst = init_expanded()

        result = inst._find_python_section_paths()

        assert result == expected

        mock_path.__truediv__.assert_called_with("PythonCook")

    @pytest.mark.parametrize(
        "shelf_exists, has_sourcefile",
        (
            (False, False),
            (True, False),
            (True, True),
        ),
    )
    def test__find_internal_shelf_item(
        self, mocker, init_expanded, shelf_exists, has_sourcefile
    ):
        """Test ExpandedOperatorType._find_internal_shelf_item."""
        mock_path = mocker.MagicMock(spec=pathlib.Path)
        mock_path.__truediv__.return_value.exists.return_value = shelf_exists

        mock_name = mocker.MagicMock(spec=str)
        mock_write_back = mocker.MagicMock(spec=bool)

        mocker.patch.object(
            houdini_package_runner.items.digital_asset.ExpandedOperatorType,
            "path",
            mock_path,
        )
        mocker.patch.object(
            houdini_package_runner.items.digital_asset.ExpandedOperatorType,
            "write_back",
            mock_write_back,
        )
        mocker.patch.object(
            houdini_package_runner.items.digital_asset.ExpandedOperatorType,
            "name",
            mock_name,
        )

        mock_source_file = mocker.MagicMock(spec=pathlib.Path)

        expected_display = (
            str(mock_source_file / "Tools.shelf") if has_sourcefile else None
        )

        mock_init_shelf = mocker.patch(
            "houdini_package_runner.items.digital_asset.ShelfFile"
        )

        inst = init_expanded()
        inst._source_file = mock_source_file if has_sourcefile else None

        result = inst._find_internal_shelf_item()

        if not shelf_exists:
            assert result is None

        else:
            assert result == mock_init_shelf.return_value

            mock_path.__truediv__.assert_called_with("Tools.shelf")

            mock_init_shelf.assert_called_with(
                mock_path.__truediv__.return_value,
                write_back=mock_write_back,
                display_name=expected_display,
                tool_name=mock_name,
            )

    @pytest.mark.parametrize(
        "shelf_exists",
        (True, False),
    )
    def test__gather_items(self, mocker, init_expanded, shelf_exists):
        """Test ExpandedOperatorType._gather_items."""
        mock_path1 = mocker.MagicMock(spec=pathlib.Path)
        mock_path2 = mocker.MagicMock(spec=pathlib.Path)
        python_paths = [mock_path1, mock_path2]
        mocker.patch.object(
            houdini_package_runner.items.digital_asset.ExpandedOperatorType,
            "_find_python_section_paths",
            return_value=python_paths,
        )

        mock_py_file1 = mocker.MagicMock(
            spec=houdini_package_runner.items.filesystem.FileToProcess
        )
        mock_py_file2 = mocker.MagicMock(
            spec=houdini_package_runner.items.filesystem.FileToProcess
        )
        python_items = [mock_py_file1, mock_py_file2]
        mock_build = mocker.patch.object(
            houdini_package_runner.items.digital_asset.ExpandedOperatorType,
            "_build_python_section_items",
            return_value=python_items,
        )

        mock_shelf_item = (
            mocker.MagicMock(spec=houdini_package_runner.items.xml.ShelfFile)
            if shelf_exists
            else None
        )
        mocker.patch.object(
            houdini_package_runner.items.digital_asset.ExpandedOperatorType,
            "_find_internal_shelf_item",
            return_value=mock_shelf_item,
        )

        mock_ds_item = mocker.MagicMock(
            spec=houdini_package_runner.items.dialog_script.DialogScriptItem
        )
        mocker.patch.object(
            houdini_package_runner.items.digital_asset.ExpandedOperatorType,
            "_get_dialog_script_item",
            return_value=mock_ds_item,
        )

        expected = [mock_py_file1, mock_py_file2]

        if shelf_exists:
            expected.append(mock_shelf_item)

        expected.append(mock_ds_item)

        inst = init_expanded()

        result = inst._gather_items()

        assert result == expected

        mock_build.assert_called_with(python_paths)

    @pytest.mark.parametrize("has_sourcefile", (True, False))
    def test__get_dialog_script_item(self, mocker, init_expanded, has_sourcefile):
        """Test ExpandedOperatorType._get_dialog_script_item."""
        mock_path = mocker.MagicMock(spec=pathlib.Path)
        mocker.patch.object(
            houdini_package_runner.items.digital_asset.ExpandedOperatorType,
            "path",
            mock_path,
        )

        mock_write_back = mocker.MagicMock(spec=bool)
        mocker.patch.object(
            houdini_package_runner.items.digital_asset.ExpandedOperatorType,
            "write_back",
            mock_write_back,
        )

        mock_init_ds = mocker.patch(
            "houdini_package_runner.items.digital_asset.DialogScriptItem"
        )

        mocker.patch.object(
            houdini_package_runner.items.digital_asset.ExpandedOperatorType,
            "name",
            "some::path/item",
        )

        inst = init_expanded()
        inst._source_file = (
            pathlib.Path("/some/path/some::source_name") if has_sourcefile else None
        )

        result = inst._get_dialog_script_item()

        assert result == mock_init_ds.return_value

        mock_path.__truediv__.assert_called_with("DialogScript")

        expected_display = (
            "some__source_name_DialogScript"
            if has_sourcefile
            else "some__path_item_DialogScript"
        )
        mock_init_ds.assert_called_with(
            mock_path.__truediv__.return_value,
            expected_display,
            write_back=mock_write_back,
        )

    @pytest.mark.parametrize("options_exist", (True, False))
    def test__get_extra_python_section_files(
        self, mocker, shared_datadir, init_expanded, options_exist
    ):
        """Test ExpandedOperatorType._get_extra_python_section_files."""
        if options_exist:
            path = shared_datadir / "test__get_extra_python_section_files"

        else:
            path = shared_datadir / "does_not_exists"

        mocker.patch.object(
            houdini_package_runner.items.digital_asset.ExpandedOperatorType,
            "path",
            path,
        )

        inst = init_expanded()
        result = inst._get_extra_python_section_files()

        if options_exist:
            assert result == [path / "PythonModule"]

        else:
            assert result == []

    # Properties

    def test_name(self, mocker, init_expanded):
        """Test ExpandedOperatorType.name."""
        mock_name = mocker.MagicMock(spec=str)

        inst = init_expanded()
        inst._name = mock_name

        assert inst.name == mock_name

        with pytest.raises(AttributeError):
            inst.name = "some name"

    # Methods

    @pytest.mark.parametrize(
        "contents_changed, return_codes, expected",
        (
            (True, (0, 0, 0), 0),
            (False, (1, 0, 0), 1),
            (False, (0, 0, 1), 1),
        ),
    )
    def test_process(
        self, mocker, init_expanded, contents_changed, return_codes, expected
    ):
        """Test ExpandedOperatorType.process."""
        mock_runner = mocker.MagicMock(
            spec=houdini_package_runner.runners.base.HoudiniPackageRunner
        )

        mock_item1 = mocker.MagicMock(
            spec=houdini_package_runner.items.filesystem.FileToProcess
        )
        mock_item1.contents_changed = False
        mock_item1.process.return_value = return_codes[0]

        mock_item2 = mocker.MagicMock(
            spec=houdini_package_runner.items.filesystem.FileToProcess
        )
        mock_item2.contents_changed = contents_changed
        mock_item2.process.return_value = return_codes[1]

        mock_item3 = mocker.MagicMock(
            spec=houdini_package_runner.items.filesystem.FileToProcess
        )
        mock_item3.contents_changed = False
        mock_item3.process.return_value = return_codes[2]

        items = [mock_item1, mock_item2, mock_item3]

        mocker.patch.object(
            houdini_package_runner.items.digital_asset.ExpandedOperatorType,
            "_gather_items",
            return_value=items,
        )

        inst = init_expanded()
        inst._contents_changed = False

        result = inst.process(mock_runner)

        assert result == expected

        assert inst._contents_changed == contents_changed

        for item in items:
            item.process.assert_called_with(mock_runner)


class TestDigitalAssetDirectory:
    """Test houdini_package_runner.items.digital_asset.DigitalAssetDirectory."""

    @pytest.mark.parametrize("has_source_file", (False, True))
    def test___init__(self, mocker, has_source_file):
        """Test object initialization."""
        mock_path = mocker.MagicMock(spec=pathlib.Path)
        mock_source_file = mocker.MagicMock(spec=pathlib.PurePath)

        mock_super_init = mocker.patch.object(
            houdini_package_runner.items.base.BaseFileItem, "__init__"
        )

        if has_source_file:
            inst = houdini_package_runner.items.digital_asset.DigitalAssetDirectory(
                mock_path, write_back=True, source_file=mock_source_file
            )

            mock_super_init.assert_called_with(mock_path, write_back=True)

            assert inst._source_file == mock_source_file

        else:
            inst = houdini_package_runner.items.digital_asset.DigitalAssetDirectory(
                mock_path
            )

            mock_super_init.assert_called_with(mock_path, write_back=False)

            assert inst._source_file is None

    @pytest.mark.parametrize("has_source_file", (True, False))
    def test__build_operator_list(self, mocker, init_asset_dir, has_source_file):
        """Test DigitalAssetDirectory._build_operator_list."""
        mock_dir_name1 = mocker.MagicMock(spec=str)
        mock_op_name1 = mocker.MagicMock(spec=str)

        mock_dir_name2 = mocker.MagicMock(spec=str)
        mock_op_name2 = mocker.MagicMock(spec=str)

        dirs = [[mock_dir_name1, mock_op_name1], [mock_dir_name2, mock_op_name2]]

        mocker.patch.object(
            houdini_package_runner.items.digital_asset.DigitalAssetDirectory,
            "_find_operator_dirs",
            return_value=dirs,
        )

        mock_path = mocker.MagicMock(spec=pathlib.Path)
        mocker.patch.object(
            houdini_package_runner.items.digital_asset.DigitalAssetDirectory,
            "path",
            mock_path,
        )

        mock_write_back = mocker.MagicMock(spec=bool)
        mocker.patch.object(
            houdini_package_runner.items.digital_asset.DigitalAssetDirectory,
            "write_back",
            mock_write_back,
        )

        mock_init_expanded = mocker.patch(
            "houdini_package_runner.items.digital_asset.ExpandedOperatorType"
        )

        mock_source_file = mocker.MagicMock(spec=pathlib.PurePath)

        inst = init_asset_dir()
        inst._source_file = mock_source_file if has_source_file else None

        if has_source_file:
            expected_display_names = [
                pathlib.PurePath(f"{mock_source_file}?{mock_op_name1}"),
                pathlib.PurePath(f"{mock_source_file}?{mock_op_name2}"),
            ]
        else:
            expected_display_names = [None, None]

        result = inst._build_operator_list()

        assert len(result) == 2

        mock_init_expanded.assert_has_calls(
            [
                mocker.call(
                    mock_path / mock_dir_name1,
                    mock_op_name1,
                    write_back=mock_write_back,
                    source_file=expected_display_names[0],
                ),
                mocker.call(
                    mock_path / mock_dir_name2,
                    mock_op_name2,
                    write_back=mock_write_back,
                    source_file=expected_display_names[1],
                ),
            ]
        )

    @pytest.mark.parametrize("section_list_exists", (True, False))
    def test__find_operator_dirs(
        self, mocker, shared_datadir, init_asset_dir, section_list_exists
    ):
        """Test DigitalAssetDirectory._find_operator_dirs."""
        path = (
            shared_datadir / "test__find_operator_dirs"
            if section_list_exists
            else shared_datadir
        )

        mocker.patch.object(
            houdini_package_runner.items.digital_asset.DigitalAssetDirectory,
            "path",
            path,
        )

        inst = init_asset_dir()

        if not section_list_exists:
            with pytest.raises(RuntimeError):
                inst._find_operator_dirs()

        else:
            result = inst._find_operator_dirs()

            assert result == [
                [
                    "com.houdinitoolbox_8_8Sop_1cppwrangle_8_81",
                    "com.houdinitoolbox::Sop/cppwrangle::1",
                ]
            ]

    # Methods

    @pytest.mark.parametrize(
        "contents_changed, return_codes, expected",
        (
            (True, (0, 0, 0), 0),
            (False, (0, 1, 0), 1),
        ),
    )
    def test_process(
        self, mocker, init_asset_dir, contents_changed, return_codes, expected
    ):
        """Test DigitalAssetDirectory.process."""
        mock_runner = mocker.MagicMock(
            spec=houdini_package_runner.runners.base.HoudiniPackageRunner
        )

        mock_operator1 = mocker.MagicMock(
            spec=houdini_package_runner.items.digital_asset.ExpandedOperatorType
        )
        mock_operator1.contents_changed = False
        mock_operator1.process.return_value = return_codes[0]

        mock_operator2 = mocker.MagicMock(
            spec=houdini_package_runner.items.digital_asset.ExpandedOperatorType
        )
        mock_operator2.contents_changed = contents_changed
        mock_operator2.process.return_value = return_codes[1]

        mock_operator3 = mocker.MagicMock(
            spec=houdini_package_runner.items.digital_asset.ExpandedOperatorType
        )
        mock_operator3.contents_changed = False
        mock_operator3.process.return_value = return_codes[2]

        operators = [mock_operator1, mock_operator2, mock_operator3]

        mocker.patch.object(
            houdini_package_runner.items.digital_asset.DigitalAssetDirectory,
            "_build_operator_list",
            return_value=operators,
        )

        inst = init_asset_dir()
        inst._contents_changed = False

        result = inst.process(mock_runner)

        assert result == expected

        assert inst._contents_changed == contents_changed

        for operator in operators:
            operator.process.assert_called_with(mock_runner)


class TestBinaryDigitalAssetFile:
    """Test houdini_package_runner.items.digital_asset.BinaryDigitalAssetFile."""

    @pytest.mark.parametrize("return_code", (1, 0))
    def test__collapse_dir(self, mocker, init_binary, return_code):
        """Test BinaryDigitalAssetFile._collapse_dir"""
        mock_target = mocker.MagicMock(spec=pathlib.Path)

        mock_path = mocker.MagicMock(spec=pathlib.Path)
        mocker.patch.object(
            houdini_package_runner.items.digital_asset.BinaryDigitalAssetFile,
            "path",
            mock_path,
        )

        mock_execute = mocker.patch(
            "houdini_package_runner.utils.execute_subprocess_command",
            return_value=return_code,
        )

        inst = init_binary()

        if return_code:
            with pytest.raises(RuntimeError):
                inst._collapse_dir("hotl", mock_target)

        else:
            inst._collapse_dir("hotl", mock_target)

        mock_execute.assert_called_with(
            ["hotl", "-l", str(mock_target), str(mock_path)]
        )

    @pytest.mark.parametrize("return_code", (1, 0))
    def test__extract_file(self, mocker, init_binary, return_code):
        """Test BinaryDigitalAssetFile._extract_file"""
        mock_target = mocker.MagicMock(spec=pathlib.Path)

        mock_path = mocker.MagicMock(spec=pathlib.Path)
        mocker.patch.object(
            houdini_package_runner.items.digital_asset.BinaryDigitalAssetFile,
            "path",
            mock_path,
        )

        mock_execute = mocker.patch(
            "houdini_package_runner.utils.execute_subprocess_command",
            return_value=return_code,
        )

        inst = init_binary()

        if return_code:
            with pytest.raises(RuntimeError):
                inst._extract_file("hotl", mock_target)

        else:
            inst._extract_file("hotl", mock_target)

        mock_execute.assert_called_with(
            ["hotl", "-t", str(mock_target), str(mock_path)]
        )

    @pytest.mark.parametrize(
        "contents_changed, write_back", ((True, True), (True, False), (False, False))
    )
    def test_process(self, mocker, init_binary, contents_changed, write_back):
        """Test BinaryDigitalAssetFile.process."""
        mock_runner = mocker.MagicMock(
            spec=houdini_package_runner.runners.base.HoudiniPackageRunner
        )

        mock_init_dir = mocker.patch(
            "houdini_package_runner.items.digital_asset.DigitalAssetDirectory"
        )

        mock_extract = mocker.patch.object(
            houdini_package_runner.items.digital_asset.BinaryDigitalAssetFile,
            "_extract_file",
        )
        mock_collapse = mocker.patch.object(
            houdini_package_runner.items.digital_asset.BinaryDigitalAssetFile,
            "_collapse_dir",
        )

        mock_path = mocker.MagicMock(spec=pathlib.Path)
        mocker.patch.object(
            houdini_package_runner.items.digital_asset.BinaryDigitalAssetFile,
            "path",
            mock_path,
        )

        mocker.patch.object(
            houdini_package_runner.items.digital_asset.BinaryDigitalAssetFile,
            "write_back",
            write_back,
        )

        target_folder = mock_runner.temp_dir.__truediv__.return_value

        inst = init_binary()
        result = inst.process(mock_runner)

        assert result == mock_init_dir.return_value.process.return_value

        mock_runner.temp_dir.__truediv__.assert_called_with(mock_path.name)

        mock_extract.assert_called_with(mock_runner.hotl_command, target_folder)

        mock_init_dir.assert_called_with(
            target_folder, write_back=write_back, source_file=mock_path
        )
        mock_init_dir.return_value.process.assert_called_with(mock_runner)

        if contents_changed and write_back:
            mock_collapse.assert_called_with(mock_runner.hotl_command, target_folder)

        else:
            mock_collapse.assert_not_called()


@pytest.mark.parametrize("exists", (True, False))
def test_is_expanded_digital_asset_dir(mocker, exists):
    """Test houdini_package_runner.items.digital_asset.is_expanded_digital_asset_dir."""
    mock_path = mocker.MagicMock(spec=pathlib.Path)
    mock_path.__truediv__.return_value.is_file.return_value = exists

    result = houdini_package_runner.items.digital_asset.is_expanded_digital_asset_dir(
        mock_path
    )

    assert result == exists

    mock_path.__truediv__.assert_called_with("Sections.list")
